/*
 *
 * Copyright (C) 2009 Tomas Mazanec
 *
 * Author:   Tomas Mazanec <mazanec at utia.cas.cz>
 * Created:  Fri Aug 28 10:16:54 CEST 2009
 *
 * Application of CUDA in DSP algorithms 
 *	- Cross Ambiguity Function (CAF) implementation
 *
 */




// ///////////////////////////////////////////////////////////////////

#ifndef _CUCOMPLEX_H_
#define _CUCOMPLEX_H_


// Complex data type
typedef float2 Complex; 
static __device__ __host__ inline Complex ComplexConj(Complex);
static __device__ __host__ inline Complex ComplexAdd(Complex, Complex);
static __device__ __host__ inline Complex ComplexScale(Complex, float);
static __device__ __host__ inline Complex ComplexMul(Complex, Complex);
static __global__ void ComplexPointwiseMulAndScale(Complex*, const Complex*, int, float);


////////////////////////////////////////////////////////////////////////////////
// CUT Complex File Handling
////////////////////////////////////////////////////////////////////////////////

CUTBoolean readFileComplexF(const char *fpath, 
			    Complex *h_data,
			    const unsigned int len, 
			    CUTBoolean reverse)
{
    printf("Input file: %s\n", fpath);
    FILE *fi;

    unsigned int i=0;
    Complex a;

    if ( (fi = fopen(fpath, "r")) == NULL) {printf("IO Error\n"); return(CUTFalse);}

    char header[128];
    fgets(header,128, fi);
    
#ifdef _DEBUG
    printf("Reading:\n%s\n", header);
#endif

    for (i=0; i<len; ++i)
    {
    
        // reverse order of data
	if ( reverse == CUTTrue )
	{
    	    fscanf(fi, "%e\t%e\n", &a.x, &a.y);
	    h_data[len-(i+1)] = a;
	}
	else 
	{
	    fscanf(fi, "%e\t%e\n", &a.x, &a.y);
	    h_data[i] = a;
	}
	
	if ( (feof(fi) != 0) && (i!=(len-1)) )
	{
	    printf("File read Error.\n");
	    fclose(fi);
	    return(CUTFalse);
	}
    }
    
    fclose(fi);
    return(CUTTrue);

}

////////////////////////////////////////////////////////////////////////////////

CUTBoolean writeFileComplexF( const char *fpath,
				Complex *data,
				const unsigned int len)
{
    printf("Output file: %s\n", fpath);
    FILE *fo;
    
    unsigned int i=0;
    
    if ( (fo = fopen(fpath, "w")) == NULL) {printf("IO Error\n"); return(CUTFalse);} 

    fprintf(fo,
    "# %d of complex numbers; col1=real part; TAB;col2=imaginary part; EOL;\n",
    len);
        
    for (i=0; i<len; ++i)
    {
	if ( (fprintf(fo,"%.7e\t%.7e\n", data[i].x, data[i].y )) <= 0 )
	{
	    printf("File write Error.\n");
	    fclose(fo);
	    return(CUTFalse);
	}
    }

    fclose(fo);
    return(CUTTrue);
    
}

////////////////////////////////////////////////////////////////////////////////
// Complex operations
////////////////////////////////////////////////////////////////////////////////

// Complex conjugate
static __device__ __host__ inline Complex ComplexConj(Complex a)
{
    Complex c;
    c.x = a.x;
    c.y = -1.0 * a.y;
    return c;
}


// Complex addition
static __device__ __host__ inline Complex ComplexAdd(Complex a, Complex b)
{
    Complex c;
    c.x = a.x + b.x;
    c.y = a.y + b.y;
    return c;
}

// Complex scale
static __device__ __host__ inline Complex ComplexScale(Complex a, float s)
{
    Complex c;
    c.x = s * a.x;
    c.y = s * a.y;
    return c;
}

// Complex multiplication
static __device__ __host__ inline Complex ComplexMul(Complex a, Complex b)
{
    Complex c;
    c.x = a.x * b.x - a.y * b.y;
    c.y = a.x * b.y + a.y * b.x;
    return c;
}

// Complex pointwise multiplication
static __global__ void ComplexPointwiseMulAndScale(Complex* a, const Complex* b, int size, float scale)
{
    const int numThreads = blockDim.x * gridDim.x;
    const int threadID = blockIdx.x * blockDim.x + threadIdx.x;
    for (int i = threadID; i < size; i += numThreads)
        a[i] = ComplexScale(ComplexMul(a[i], b[i]), scale);     
} 

#endif //  _CUCOMPLEX_H_

