/*
 *
 * Copyright (C) 2009 Tomas Mazanec
 *
 * Author:   Tomas Mazanec <mazanec at utia.cas.cz>
 * Created:  Fri Aug 28 10:16:54 CEST 2009
 *
 * Application of CUDA in DSP algorithms 
 *	- Digital FIR filter implementation
 *
 */


// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <unistd.h>
#include <time.h>

// includes, project
#include <cutil_inline.h>

// includes, kernels
#include <cutesto.cu>
#include <tFIR_kernel.cu>

#define MAXTHREADSROW 512

////////////////////////////////////////////////////////////////////////////////
// declaration, forward
void runTest( int argc, char** argv);

extern "C"
void computeGold( float* reference, float* coeff, float* idata, const unsigned int M, const unsigned int x_len);

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int
main( int argc, char** argv) 
{
    runTest( argc, argv);

    cutilExit(argc, argv);
}

////////////////////////////////////////////////////////////////////////////////
//! Run a simple test for CUDA
////////////////////////////////////////////////////////////////////////////////
void
runTest( int argc, char** argv) 
{

int dev = 0;
    // use command-line specified CUDA device, otherwise use device with highest Gflop
    if( cutCheckCmdLineFlag(argc, (const char**)argv, "device") )
	cutilDeviceInit(argc, argv);
    else
	cutilSafeCall(cudaSetDevice(dev));
    

    int NB = 20; // 10000
    if (cutCheckCmdLineFlag(argc, (const char**)argv, "NB") )
	cutGetCmdLineArgumenti(argc, (const char **) argv, "NB", &NB);

    // desired FIR order // max is 3584 (=512*7) to fit to __shared__ mem
    int MD = 2048;
    if (cutCheckCmdLineFlag(argc, (const char**)argv, "M") )
	cutGetCmdLineArgumenti(argc, (const char **) argv, "M", &MD);


    unsigned int timer = 0;
    cutilCheckError( cutCreateTimer( &timer));

    const char *fpath="../../../projects/tFIR_implementationI/data/fir2DP.dat";

    // round to the nearest multiple of 512
    const unsigned int M = (const unsigned int) 
				MAXTHREADSROW * ceilf((float)MD / (float)MAXTHREADSROW);

    unsigned int K = ( (unsigned int)(4096/M) - 1 );  // max=127\M=32, max=7\M=512, max=3\M=1024, max=1\M>=2048
    if ( K < 1) { K = 1;}
    
    const unsigned int N = M*(K+1);
    const unsigned int P = N-M+1;

    //if M>512 ? NT=512: NT=M
    const unsigned int NT = ((M/MAXTHREADSROW) && 1) * MAXTHREADSROW +
			    (!(M/MAXTHREADSROW)) * M;

    const unsigned int L = (const unsigned int) 
				ceilf((float)M / (float)MAXTHREADSROW);
    
    printf("->%d-th order FIR-> M forced to: %d, No. of loops:%d, with No. of threads:%d\n",MD, M, L, NT);


    const unsigned int x_len = (NB*K+1)*M;
    const unsigned int y_len = x_len-M+1;

    printf("FIR order M: %d with NT %d threads\n", M, NT);
    printf("No. of Blocks:%d, where K:%d, perBlock inp:%d perBlock out:%d\n", NB, K, N, P);
    printf("\txlen:%d , y_len:%d\n\n", x_len, y_len);


    unsigned int i = 0;

    unsigned int x_size = sizeof(float) * x_len;
    unsigned int y_size = sizeof(float) * y_len;
    unsigned int coeff_size = sizeof(float) * M;

    // c i o -> shmem
//    unsigned int shmem_size = (P + M + N)*sizeof(float);
    // i o -> shmem
//    unsigned int shmem_size = (P + N)*sizeof(float);
    // c o -> shmem
//    unsigned int shmem_size = (P + M)*sizeof(float);
    // o -> shmem
    unsigned int shmem_size = (P)*sizeof(float);


    
    // double(2^-23) = 1.192092895507812e-07 // single(2^-23) = 1.1920929e-07
    const float epsilon = 1.1920929e-07;

    // coeffs
    float *h_coeff = (float*) malloc( coeff_size);
    // read reversed; No. of coeffs up to MD, if M>MD : h_coeff[MD..M-1] = 0.0
    cutilCheckError( readFileFCoeff(fpath, h_coeff, MD, CUTTrue) );
//    for(i = 0; i < M; i++) {printf("coeff:%e\n", h_coeff[i]);}

    //to global mem
    float* d_coeff;
    cutilSafeCall( cudaMalloc( (void**) &d_coeff, coeff_size));
//    cutilSafeCall( cudaMemcpy( d_coeff, h_coeff, coeff_size, cudaMemcpyHostToDevice) );

    // or to const mem
    cutilSafeCall( cudaMemcpyToSymbol( c_coeff, h_coeff, coeff_size) );
    // or in texture by cudaArray
    cudaChannelFormatDesc floatCTex = 
				cudaCreateChannelDesc<float>();
    cudaArray *ca_coeff;
    cutilSafeCall(cudaMallocArray(&ca_coeff, 
				    &floatCTex,
				    coeff_size, 1*sizeof(float)) );

    cutilSafeCall(cudaMemcpyToArray(ca_coeff,
				    0, 0, 
				    h_coeff, coeff_size,
				    cudaMemcpyHostToDevice) );
    
    cutilSafeCall(cudaBindTextureToArray(texCoeff, ca_coeff) );
    

    float *h_idata = (float*) malloc( x_size);
    // input data
    srand( (unsigned int)time(NULL) );

    for(i = 0; i < x_len; i++) 
    {
	if (i < (x_len-(M-MD)) ) {
    	    h_idata[i] = 
		    (float) ( ( (float)rand() - (float)RAND_MAX/2 ) / (float)RAND_MAX );
	}
	// truncate unreacheable input; if M>MD: h_coeff[MD..M-1] = 0
	else { h_idata[i] = (float)0.0;}
//	printf("inp:%e\n", h_idata[i]);
    }

    
    // host to device
    float* d_idata;
    cutilSafeCall( cudaMalloc( (void**) &d_idata, x_size));
    cutilSafeCall( cudaMemcpy( d_idata, h_idata, x_size,
                                    cudaMemcpyHostToDevice) );
    
    
    // input data in texture by cudaArray or linear memory
    cudaChannelFormatDesc floatTex = 
				cudaCreateChannelDesc<float>();
    // cuda array , dimensions of texture limited to 2^13 x 2^14
/*    cudaArray *ca_idata;
    cutilSafeCall(cudaMallocArray(&ca_idata, 
				    &floatTex,
				    x_size, 1*sizeof(float)) );
    cutilSafeCall(cudaMemcpyToArray(ca_idata,
				    0, 0, 
				    h_idata, x_size,
				    cudaMemcpyHostToDevice) );
    
    cutilSafeCall(cudaBindTextureToArray(texIData, ca_idata) );
*/

    // linear memory , dimensions of texture limited to 2^27 , 1D only
// already in  kernel.cu // texture<float, 1, cudaReadModeElementType> texIData;
    const textureReference* texIDataPtr;
    cutilSafeCall(cudaGetTextureReference(&texIDataPtr, "texIData") );

    cutilSafeCall(cudaBindTexture(0, texIDataPtr, (void *)d_idata, &floatTex, x_size) );

    
    float* d_odata;
    cutilSafeCall( cudaMalloc( (void**) &d_odata, y_size));


    // setup execution parameters
    dim3  grid( NB, 1, 1);
    dim3  threads(NT, 1, 1);

    cutilSafeCall(cudaThreadSynchronize() );
    cutilCheckError( cutStartTimer( timer));
    
/* **************************************************************************** */

    // execute the kernel
    testKernel<<< grid, threads, shmem_size >>>( d_odata, d_idata, M, N, K, L);

/* **************************************************************************** */

    // check if kernel execution generated and error // when _DEBUG also calls cudaThreadSynchronize() 
    cutilCheckMsg("Kernel execution failed");

#ifndef _DEBUG    
    // wait till kernel is done
    cutilSafeCall(cudaThreadSynchronize() );
#endif

    cutilCheckError( cutStopTimer( timer));
    float timeker = cutGetTimerValue( timer);
    cutilCheckError( cutDeleteTimer( timer));


    // result
    float* h_odata = (float*) malloc( y_size);
    cutilSafeCall( cudaMemcpy( h_odata, d_odata, y_size,
                                cudaMemcpyDeviceToHost) );



    // compute reference solution
    float* reference = (float*) malloc( y_size);
    
    if (x_len < 1e6) {
    computeGold( reference, h_coeff, h_idata, M, x_len);

    // check result
    if (x_len < 2e4) {
    cutilCheckError( writeFileF("../../../projects/tFIR_implementationI/data/input.dat",     h_idata,   x_len) );
    cutilCheckError( writeFileF("../../../projects/tFIR_implementationI/data/output.dat",    h_odata,   y_len) );
    cutilCheckError( writeFileF("../../../projects/tFIR_implementationI/data/reference.dat", reference, y_len) );
    cutilCheckError( writeFileF("../../../projects/tFIR_implementationI/data/coeffRev.dat",     h_coeff,   M) );
    }
    else {printf("Data files would be too long , skipping.\n");}

    CUTBoolean res = compareF (reference, h_odata, y_len, epsilon);
    printf( "Test %s\n", (1 == res) ? "PASSED" : "FAILED");
    }
    else {printf("Don't care about refernece , skipping.\n");}


    unsigned long num_ops = (unsigned long) (M * y_len);
    float flopsy = (num_ops/(timeker/1000)) / 1e6;
    printf( "Processing time: %f (ms) and float MULs: %e\n MegaFlopsy: %f\n\n",
	    timeker, ((float)num_ops), flopsy);



    // cleanup memory
    free( h_coeff);
    free( h_idata);
    free( h_odata);
    free( reference);
    
    cutilSafeCall( cudaUnbindTexture(texIDataPtr) );
    cutilSafeCall( cudaUnbindTexture(texCoeff) );
    cutilSafeCall(cudaFreeArray(ca_coeff));
    cutilSafeCall(cudaFree(d_odata));
    cutilSafeCall(cudaFree(d_idata));
    
    cudaThreadExit();
    
}
